# -*- coding: utf-8 -*-
#
# Copyright 2013 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import pkg_resources as pkg
import sys
import tempfile

from oslo.config import cfg
from oslo import messaging
from oslo.messaging import transport
from oslotest import base
from stevedore import driver
import testtools.matchers as ttm

from mistral import context as auth_context
from mistral.db.sqlalchemy import api as db_api
from mistral import engine
from mistral.engine import executor
from mistral.openstack.common import log as logging
from mistral import version


RESOURCES_PATH = 'tests/resources/'
LOG = logging.getLogger(__name__)


def get_resource(resource_name):
    return open(pkg.resource_filename(
        version.version_info.package,
        RESOURCES_PATH + resource_name)).read()


def get_fake_transport():
    # Get transport here to let oslo.messaging setup default config
    # before changing the rpc_backend to the fake driver; otherwise,
    # oslo.messaging will throw exception.
    messaging.get_transport(cfg.CONF)
    cfg.CONF.set_default('rpc_backend', 'fake')
    url = transport.TransportURL.parse(cfg.CONF, None, None)
    kwargs = dict(default_exchange=cfg.CONF.control_exchange,
                  allowed_remote_exmods=[])
    mgr = driver.DriverManager('oslo.messaging.drivers',
                               url.transport,
                               invoke_on_load=True,
                               invoke_args=[cfg.CONF, url],
                               invoke_kwds=kwargs)
    return transport.Transport(mgr.driver)


class BaseTest(base.BaseTestCase):
    def assertListEqual(self, l1, l2):
        if tuple(sys.version_info)[0:2] < (2, 7):
            # for python 2.6 compatibility
            self.assertEqual(l1, l2)
        else:
            super(BaseTest, self).assertListEqual(l1, l2)

    def assertDictEqual(self, cmp1, cmp2):
        if tuple(sys.version_info)[0:2] < (2, 7):
            # for python 2.6 compatibility
            self.assertThat(cmp1, ttm.Equals(cmp2))
        else:
            super(BaseTest, self).assertDictEqual(cmp1, cmp2)

    def _assert_single_item(self, items, **props):
        return self._assert_multiple_items(items, 1, **props)[0]

    def _assert_multiple_items(self, items, count, **props):
        def _matches(item, **props):
            for prop_name, prop_val in props.iteritems():
                v = item[prop_name] if isinstance(item, dict) \
                    else getattr(item, prop_name)

                if v != prop_val:
                    return False

            return True

        filtered_items = filter(lambda item: _matches(item, **props), items)

        found = len(filtered_items)

        if found != count:
            LOG.info("[failed test ctx] items=%s, expected_props=%s" % (str(
                items), props))
            self.fail("Wrong number of items found [props=%s, "
                      "expected=%s, found=%s]" % (props, count, found))

        return filtered_items


class DbTestCase(BaseTest):
    def setUp(self):
        super(DbTestCase, self).setUp()
        _db_fd, self.db_path = tempfile.mkstemp()
        cfg.CONF.set_default('connection', 'sqlite:///' + self.db_path,
                             group='database')
        db_api.setup_db()
        self.addCleanup(db_api.drop_db)

        self.ctx = auth_context.MistralContext(user_id='1-2-3-4',
                                               project_id='5-6-7-8',
                                               user_name='test-user',
                                               project_name='test-project',
                                               is_admin=False)
        auth_context.set_ctx(self.ctx)
        self.addCleanup(auth_context.set_ctx, None)

    def is_db_session_open(self):
        return db_api._get_thread_local_session() is not None


class EngineTestCase(DbTestCase):
    transport = get_fake_transport()
    backend = engine.get_engine(cfg.CONF.engine.engine, transport)

    def __init__(self, *args, **kwargs):
        super(EngineTestCase, self).__init__(*args, **kwargs)
        self.engine = engine.EngineClient(self.transport)

    @classmethod
    def mock_task_result(cls, task_id, state, result):
        """Mock the engine convey_task_results to send request directly
        to the engine instead of going through the oslo.messaging transport.
        """
        cntx = {}
        kwargs = {'task_id': task_id,
                  'state': state,
                  'result': result}
        return cls.backend.convey_task_result(cntx, **kwargs)

    @classmethod
    def mock_start_workflow(cls, workbook_name, task_name, context=None):
        """Mock the engine start_workflow_execution to send request directly
        to the engine instead of going through the oslo.messaging transport.
        """
        cntx = {}
        kwargs = {'workbook_name': workbook_name,
                  'task_name': task_name,
                  'context': context}
        return cls.backend.start_workflow_execution(cntx, **kwargs)

    @classmethod
    def mock_get_workflow_state(cls, workbook_name, execution_id):
        """Mock the engine get_workflow_execution_state to send request
        directly to the engine instead of going through the oslo.messaging
        transport.
        """
        cntx = {}
        kwargs = {'workbook_name': workbook_name,
                  'execution_id': execution_id}
        return cls.backend.get_workflow_execution_state(cntx, **kwargs)

    @classmethod
    def mock_run_tasks(cls, tasks):
        """Mock the engine _run_tasks to send requests directly to the task
        executor instead of going through the oslo.messaging transport.
        """
        exctr = executor.get_executor(cfg.CONF.engine.engine, cls.transport)
        for task in tasks:
            exctr.handle_task({}, task=task)

    @classmethod
    def mock_handle_task(cls, cntx, **kwargs):
        """Mock the executor handle_task to send requests directory to the task
        executor instead of going through the oslo.messaging transport.
        """
        exctr = executor.get_executor(cfg.CONF.engine.engine, cls.transport)
        return exctr.handle_task(cntx, **kwargs)



# -*- coding: utf-8 -*-
#
# Copyright 2014 - StackStorm, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import eventlet
import mock
from oslo.config import cfg

from mistral.actions import std_actions
from mistral.db import api as db_api
from mistral import dsl_parser as parser
from mistral import engine
from mistral.engine.drivers.default import engine as concrete_engine
from mistral.engine import states
from mistral import exceptions as exc
from mistral.openstack.common import log as logging
from mistral.tests import base


LOG = logging.getLogger(__name__)

# Use the set_default method to set value otherwise in certain test cases
# the change in value is not permanent.
cfg.CONF.set_default('auth_enable', False, group='pecan')

WB_NAME = "my_workbook"


def _get_workbook(workbook_name):
    wb = db_api.workbook_get(workbook_name)
    return parser.get_workbook(wb["definition"])


class FailBeforeSuccessMocker(object):
    def __init__(self, fail_count):
        self._max_fail_count = fail_count
        self._call_count = 0

    def mock_partial_failure(self, *args):
        if self._call_count < self._max_fail_count:
            self._call_count += 1
            raise exc.ActionException()

        return "result"


@mock.patch.object(
    engine.EngineClient, 'start_workflow_execution',
    mock.MagicMock(side_effect=base.EngineTestCase.mock_start_workflow))
@mock.patch.object(
    engine.EngineClient, 'convey_task_result',
    mock.MagicMock(side_effect=base.EngineTestCase.mock_task_result))
@mock.patch.object(
    concrete_engine.DefaultEngine, '_run_tasks',
    mock.MagicMock(side_effect=base.EngineTestCase.mock_run_tasks))
@mock.patch.object(
    std_actions.HTTPAction, 'run',
    mock.MagicMock(return_value='result'))
class TaskRetryTest(base.EngineTestCase):

    @mock.patch.object(
        db_api, 'workbook_get',
        mock.MagicMock(return_value={
            'definition': base.get_resource('retry_task/retry_task.yaml')}))
    def test_no_retry(self):
        execution = self.engine.start_workflow_execution(WB_NAME,
                                                         'retry_task', None)
        tasks = db_api.tasks_get(workbook_name=WB_NAME,
                                 execution_id=execution['id'])

        self.engine.convey_task_result(tasks[0]['id'], states.SUCCESS,
                                       {'output': 'result'})

        # TODO(rakhmerov): It's not stable, need to avoid race condition.
        self._assert_single_item(tasks, name='retry_task')
        self._assert_single_item(tasks, task_runtime_context=None)

    @mock.patch.object(
        db_api, 'workbook_get',
        mock.MagicMock(return_value={
            'definition': base.get_resource('retry_task/retry_task.yaml')}))
    def test_retry_always_error(self):
        workbook = _get_workbook(WB_NAME)

        execution = self.engine.start_workflow_execution(WB_NAME,
                                                         'retry_task', None)
        tasks = db_api.tasks_get(workbook_name=WB_NAME,
                                 execution_id=execution['id'])
        task_spec = workbook.tasks.get(tasks[0]['name'])
        retry_count, _, __ = task_spec.get_retry_parameters()

        for x in xrange(0, retry_count + 1):
            self.engine.convey_task_result(tasks[0]['id'], states.ERROR,
                                           {'output': 'result'})

        # TODO(rakhmerov): It's not stable, need to avoid race condition.
        tasks = db_api.tasks_get(workbook_name=WB_NAME,
                                 execution_id=execution['id'])

        self._assert_single_item(tasks, name='retry_task')
        self._assert_single_item(tasks, task_runtime_context={
            'retry_no': retry_count - 1})
        self._assert_single_item(tasks, state=states.ERROR)

    @mock.patch.object(
        db_api, 'workbook_get',
        mock.MagicMock(return_value={
            'definition': base.get_resource('retry_task/retry_task.yaml')}))
    def test_retry_eventual_success(self):
        workbook = _get_workbook(WB_NAME)

        execution = self.engine.start_workflow_execution(WB_NAME,
                                                         'retry_task', None)
        tasks = db_api.tasks_get(workbook_name=WB_NAME,
                                 execution_id=execution['id'])
        task_spec = workbook.tasks.get(tasks[0]['name'])
        retry_count, _, __ = task_spec.get_retry_parameters()

        for x in xrange(0, retry_count / 2):
            self.engine.convey_task_result(tasks[0]['id'], states.ERROR,
                                           {'output': 'result'})

        self.engine.convey_task_result(tasks[0]['id'], states.SUCCESS,
                                       {'output': 'result'})

        # TODO(rakhmerov): It's not stable, need to avoid race condition.
        tasks = db_api.tasks_get(workbook_name=WB_NAME,
                                 execution_id=execution['id'])

        self._assert_single_item(tasks, name='retry_task')
        self._assert_single_item(tasks, task_runtime_context={
            'retry_no': retry_count / 2 - 1})

    @mock.patch.object(
        db_api, 'workbook_get',
        mock.MagicMock(return_value={'definition': base.get_resource(
            'retry_task/delay_retry_task.yaml')}))
    def test_retry_delay(self):
        task_name = 'delay_retry_task'
        workbook = _get_workbook(WB_NAME)

        execution = self.engine.start_workflow_execution(WB_NAME,
                                                         task_name, None)

        tasks = db_api.tasks_get(workbook_name=WB_NAME,
                                 execution_id=execution['id'])
        task_spec = workbook.tasks.get(tasks[0]['name'])
        retry_count, _, delay = task_spec.get_retry_parameters()

        for x in xrange(0, retry_count):
            self.engine.convey_task_result(tasks[0]['id'], states.ERROR,
                                           {'output': 'result'})

            tasks = db_api.tasks_get(workbook_name=WB_NAME,
                                     execution_id=execution['id'])

            # TODO(rakhmerov): It's not stable, need to avoid race condition.
            self._assert_single_item(tasks, name=task_name)
            self._assert_single_item(tasks, state=states.DELAYED)

            eventlet.sleep(delay * 2)

        # Convey final result outside the loop.
        self.engine.convey_task_result(tasks[0]['id'], states.ERROR,
                                       {'output': 'result'})

        # TODO(rakhmerov): It's not stable, need to avoid race condition.
        tasks = db_api.tasks_get(workbook_name=WB_NAME,
                                 execution_id=execution['id'])

        self._assert_single_item(tasks, name=task_name)
        self._assert_single_item(tasks, task_runtime_context={
            'retry_no': retry_count - 1})
        self._assert_single_item(tasks, state=states.ERROR)

    @mock.patch.object(
        db_api, 'workbook_get',
        mock.MagicMock(return_value={
            'definition': base.get_resource('retry_task/two_tasks.yaml')}))
    def test_from_no_retry_to_retry_task(self):
        task_name_1 = 'no_retry_task'
        task_name_2 = 'delay_retry_task'
        workbook = _get_workbook(WB_NAME)

        execution = self.engine.start_workflow_execution(WB_NAME,
                                                         task_name_1, None)

        tasks = db_api.tasks_get(workbook_name=WB_NAME,
                                 execution_id=execution['id'])

        self._assert_single_item(tasks, name=task_name_1)

        self.engine.convey_task_result(tasks[0]['id'], states.SUCCESS,
                                       {'output': 'result'})

        tasks = db_api.tasks_get(workbook_name=WB_NAME,
                                 execution_id=execution['id'])

        self._assert_single_item(tasks, name=task_name_2)

        task_spec = workbook.tasks.get(task_name_2)
        retry_count, _, delay = task_spec.get_retry_parameters()

        for x in xrange(0, retry_count):
            self.engine.convey_task_result(tasks[1]['id'], states.ERROR,
                                           {'output': 'result'})

            tasks = db_api.tasks_get(workbook_name=WB_NAME,
                                     execution_id=execution['id'])

            # TODO(rakhmerov): It's not stable, need to avoid race condition.
            self._assert_single_item(tasks, name=task_name_1)
            self._assert_single_item(tasks, state=states.DELAYED)

            eventlet.sleep(delay * 2)

        # Convey final result outside the loop.
        self.engine.convey_task_result(tasks[1]['id'], states.ERROR,
                                       {'output': 'result'})

        # TODO(rakhmerov): It's not stable, need to avoid race condition.
        tasks = db_api.tasks_get(workbook_name=WB_NAME,
                                 execution_id=execution['id'])

        self._assert_single_item(tasks, name=task_name_2)
        self._assert_single_item(tasks, task_runtime_context={
            'retry_no': retry_count - 1})
        self._assert_single_item(tasks, state=states.ERROR)

    @mock.patch.object(std_actions.EchoAction, "run",
                       mock.MagicMock(side_effect=exc.ActionException))
    @mock.patch.object(
        db_api, 'workbook_get',
        mock.MagicMock(return_value={
            'definition': base.get_resource('retry_task/sync_task.yaml')}))
    def test_sync_action_always_error(self):
        workbook = _get_workbook(WB_NAME)
        start_task = 'sync-task'
        task_spec = workbook.tasks.get(start_task)
        retry_count, _, __ = task_spec.get_retry_parameters()

        execution = self.engine.start_workflow_execution(WB_NAME,
                                                         start_task, None)

        # TODO(rakhmerov): It's not stable, need to avoid race condition.
        tasks = db_api.tasks_get(workbook_name=WB_NAME,
                                 execution_id=execution['id'])

        self._assert_single_item(tasks, name=start_task)
        self._assert_single_item(tasks, task_runtime_context={
            'retry_no': retry_count - 1})
        self._assert_single_item(tasks, state=states.ERROR)

    @mock.patch.object(
        db_api, 'workbook_get',
        mock.MagicMock(return_value={
            'definition': base.get_resource('retry_task/sync_task.yaml')}))
    def test_sync_action_eventual_success(self):
        start_task = 'sync-task'
        workbook = _get_workbook(WB_NAME)
        task_spec = workbook.tasks.get(start_task)
        retry_count, _, __ = task_spec.get_retry_parameters()

        # After a pre-set no of retries the mock method will return a
        # success to simulate this test-case.
        mock_functor = FailBeforeSuccessMocker(retry_count / 2 + 1)

        with mock.patch.object(std_actions.EchoAction, "run",
                               side_effect=mock_functor.mock_partial_failure):
            execution = self.engine.start_workflow_execution(WB_NAME,
                                                             start_task,
                                                             None)

            # TODO(rakhmerov): It's not stable, need to avoid race condition.
            tasks = db_api.tasks_get(workbook_name=WB_NAME,
                                     execution_id=execution['id'])

            self._assert_single_item(tasks, name=start_task)
            self._assert_single_item(tasks, task_runtime_context={
                'retry_no': retry_count / 2})
            self._assert_single_item(tasks, state=states.SUCCESS)



# -*- coding: utf-8 -*-
#
# Copyright 2013 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import json
import mock
from webtest import app as webtest_app

from mistral.db import api as db_api
from mistral import engine
from mistral import exceptions as ex
from mistral.tests.api import base

# TODO(everyone): later we need additional tests verifying all the errors etc.


EXECS = [
    {
        'id': '123',
        'workbook_name': 'my_workbook',
        'task': 'my_task',
        'state': 'RUNNING',
        'context': {
            "person": {
                "first_name": "John",
                "last_name": "Doe"
            }
        }
    }
]


WORKBOOKS = [
    {
        'name': "my_workbook",
        'description': "My cool Mistral workbook",
        'tags': ['deployment', 'demo']
    }
]


UPDATED_EXEC = EXECS[0].copy()
UPDATED_EXEC['state'] = 'STOPPED'


def canonize(json_dict):
    if json_dict.get('context'):
        json_dict['context'] = json.loads(json_dict['context'])

    return json_dict


class TestExecutionsController(base.FunctionalTest):
    @mock.patch.object(db_api, 'execution_get',
                       mock.MagicMock(return_value=EXECS[0]))
    def test_workbook_get(self):
        resp = self.app.get('/v1/workbooks/my_workbook/executions/123')

        self.assertEqual(resp.status_int, 200)
        self.assertDictEqual(EXECS[0], canonize(resp.json))

    @mock.patch.object(db_api, 'execution_get',
                       mock.MagicMock(return_value=EXECS[0]))
    def test_root_get(self):
        resp = self.app.get('/v1/executions/123')

        self.assertEqual(resp.status_int, 200)
        self.assertDictEqual(EXECS[0], canonize(resp.json))

    @mock.patch.object(db_api, 'execution_get',
                       mock.MagicMock(side_effect=ex.NotFoundException()))
    def test_get_empty(self):
        resp = self.app.get('/v1/workbooks/my_workbook/executions/123',
                            expect_errors=True)
        self.assertEqual(resp.status_int, 404)

    @mock.patch.object(db_api, 'execution_update',
                       mock.MagicMock(return_value=UPDATED_EXEC))
    def test_workbook_put(self):
        resp = self.app.put_json('/v1/workbooks/my_workbook/executions/123',
                                 dict(state='STOPPED'))

        self.assertEqual(resp.status_int, 200)
        self.assertDictEqual(UPDATED_EXEC, canonize(resp.json))

    @mock.patch.object(db_api, 'execution_update',
                       mock.MagicMock(return_value=UPDATED_EXEC))
    def test_root_put(self):
        resp = self.app.put_json('/v1/executions/123',
                                 dict(state='STOPPED'))

        self.assertEqual(resp.status_int, 200)
        self.assertDictEqual(UPDATED_EXEC, canonize(resp.json))

    @mock.patch.object(db_api, 'execution_update',
                       mock.MagicMock(
                           side_effect=ex.NotFoundException()))
    def test_put_not_found(self):
        resp = self.app.put_json('/v1/workbooks/my_workbook/executions/123',
                                 dict(state='STOPPED'), expect_errors=True)

        self.assertEqual(resp.status_int, 404)

    @mock.patch.object(engine.EngineClient, 'start_workflow_execution',
                       mock.MagicMock(return_value=EXECS[0]))
    @mock.patch.object(db_api, 'workbook_definition_get',
                       mock.Mock(return_value="Workflow:"))
    def test_post(self):
        my_workbook = WORKBOOKS[0]
        self.app.post_json('/v1/workbooks',
                           my_workbook)

        new_exec = EXECS[0].copy()
        new_exec['context'] = json.dumps(new_exec['context'])

        resp = self.app.post_json('/v1/workbooks/my_workbook/executions',
                                  new_exec)
        self.assertEqual(resp.status_int, 201)
        self.assertDictEqual(EXECS[0], canonize(resp.json))

    @mock.patch.object(engine.EngineClient, 'start_workflow_execution',
                       mock.MagicMock(return_value=EXECS[0]))
    def test_post_definition_empty(self):
        my_workbook = WORKBOOKS[0]
        self.app.post_json('/v1/workbooks',
                           my_workbook)

        new_exec = EXECS[0].copy()
        new_exec['context'] = json.dumps(new_exec['context'])

        resp = self.app.post_json('/v1/workbooks/my_workbook/executions',
                                  new_exec, expect_errors=True)
        self.assertEqual(resp.status_int, 404)

    @mock.patch.object(engine.EngineClient, 'start_workflow_execution',
                       mock.MagicMock(side_effect=ex.MistralException))
    def test_post_throws_exception(self):
        context = self.assertRaises(webtest_app.AppError, self.app.post_json,
                                    '/v1/workbooks/my_workbook/executions',
                                    EXECS[0])
        self.assertIn('Bad response: 400', context.message)

    @mock.patch.object(db_api, 'execution_delete',
                       mock.MagicMock(return_value=None))
    def test_workbook_delete(self):
        resp = self.app.delete('/v1/workbooks/my_workbook/executions/123')

        self.assertEqual(resp.status_int, 204)

    @mock.patch.object(db_api, 'execution_delete',
                       mock.MagicMock(return_value=None))
    def test_root_delete(self):
        resp = self.app.delete('/v1/executions/123')

        self.assertEqual(resp.status_int, 204)

    @mock.patch.object(db_api, 'execution_delete',
                       mock.MagicMock(side_effect=ex.NotFoundException))
    def test_delete_not_found(self):
        resp = self.app.delete('/v1/workbooks/my_workbook/executions/123',
                               expect_errors=True)

        self.assertEqual(resp.status_int, 404)

    @mock.patch.object(db_api, 'executions_get',
                       mock.MagicMock(return_value=EXECS))
    @mock.patch.object(db_api, 'workbook_get',
                       mock.MagicMock(return_value={'name': 'my_workbook'}))
    def test_workbook_get_all(self):
        resp = self.app.get('/v1/workbooks/my_workbook/executions')

        self.assertEqual(resp.status_int, 200)

        self.assertEqual(len(resp.json), 1)
        self.assertDictEqual(EXECS[0], canonize(resp.json['executions'][0]))

    @mock.patch.object(db_api, 'executions_get',
                       mock.MagicMock(return_value=EXECS))
    @mock.patch.object(db_api, 'workbook_get',
                       mock.MagicMock(return_value={'name': 'my_workbook'}))
    def test_root_get_all(self):
        resp = self.app.get('/v1/executions')

        self.assertEqual(resp.status_int, 200)

        self.assertEqual(len(resp.json), 1)
        self.assertDictEqual(EXECS[0], canonize(resp.json['executions'][0]))

    @mock.patch.object(db_api, 'executions_get',
                       mock.MagicMock(return_value=EXECS))
    def test_get_all_no_workbook(self):
        resp = self.app.get('/v1/workbooks/my_workbook/executions',
                            expect_errors=True)

        self.assertEqual(resp.status_int, 404)



# Copyright 2011 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import contextlib
import errno
import os
import tempfile

from mistral.openstack.common import excutils
from mistral.openstack.common import log as logging

LOG = logging.getLogger(__name__)

_FILE_CACHE = {}


def ensure_tree(path):
    """Create a directory (and any ancestor directories required)

    :param path: Directory to create
    """
    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno == errno.EEXIST:
            if not os.path.isdir(path):
                raise
        else:
            raise


def read_cached_file(filename, force_reload=False):
    """Read from a file if it has been modified.

    :param force_reload: Whether to reload the file.
    :returns: A tuple with a boolean specifying if the data is fresh
              or not.
    """
    global _FILE_CACHE

    if force_reload and filename in _FILE_CACHE:
        del _FILE_CACHE[filename]

    reloaded = False
    mtime = os.path.getmtime(filename)
    cache_info = _FILE_CACHE.setdefault(filename, {})

    if not cache_info or mtime > cache_info.get('mtime', 0):
        LOG.debug("Reloading cached file %s" % filename)
        with open(filename) as fap:
            cache_info['data'] = fap.read()
        cache_info['mtime'] = mtime
        reloaded = True
    return (reloaded, cache_info['data'])


def delete_if_exists(path, remove=os.unlink):
    """Delete a file, but ignore file not found error.

    :param path: File to delete
    :param remove: Optional function to remove passed path
    """

    try:
        remove(path)
    except OSError as e:
        if e.errno != errno.ENOENT:
            raise


@contextlib.contextmanager
def remove_path_on_error(path, remove=delete_if_exists):
    """Protect code that wants to operate on PATH atomically.
    Any exception will cause PATH to be removed.

    :param path: File to work with
    :param remove: Optional function to remove passed path
    """

    try:
        yield
    except Exception:
        with excutils.save_and_reraise_exception():
            remove(path)


def file_open(*args, **kwargs):
    """Open file

    see built-in file() documentation for more details

    Note: The reason this is kept in a separate module is to easily
    be able to provide a stub module that doesn't alter system
    state at all (for unit tests)
    """
    return open(*args, **kwargs)


def write_to_tempfile(content, path=None, suffix='', prefix='tmp'):
    """Create temporary file or use existing file.

    This util is needed for creating temporary file with
    specified content, suffix and prefix. If path is not None,
    it will be used for writing content. If the path doesn't
    exist it'll be created.

    :param content: content for temporary file.
    :param path: same as parameter 'dir' for mkstemp
    :param suffix: same as parameter 'suffix' for mkstemp
    :param prefix: same as parameter 'prefix' for mkstemp

    For example: it can be used in database tests for creating
    configuration files.
    """
    if path:
        ensure_tree(path)

    (fd, path) = tempfile.mkstemp(suffix=suffix, dir=path, prefix=prefix)
    try:
        os.write(fd, content)
    finally:
        os.close(fd)
    return path



# -*- coding: utf-8 -*-
#
# Copyright 2013 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from mistral.db import api as db_api
from mistral.engine import data_flow
from mistral.openstack.common import log as logging
from mistral.tests import base

LOG = logging.getLogger(__name__)

WB_NAME = 'my_workbook'
EXEC_ID = '1'

CONTEXT = {
    'param1': 'val1',
    'param2': 'val2',
    'param3': {
        'param31': 'val31',
        'param32': 'val32'
    }
}

TASK = {
    'workbook_name': WB_NAME,
    'execution_id': EXEC_ID,
    'name': 'my_task',
    'task_spec': {
        'parameters': {
            'p1': 'My string',
            'p2': '$.param3.param32',
            'p3': ''
        },
        'publish': {
            'new_key11': '$.new_key1'
        }
    },
    'service_spec': {
        'actions': {
            'action': {
                'output': {
                    # This one should not be evaluated.
                    'server_id': '$.server.id'
                }
            }
        }
    },
    'in_context': CONTEXT
}


class DataFlowModuleTest(base.DbTestCase):
    def test_evaluate_task_parameters(self):
        parameters = data_flow.evaluate_task_parameters(TASK, CONTEXT)

        self.assertEqual(3, len(parameters))
        self.assertEqual('My string', parameters['p1'])
        self.assertEqual('val32', parameters['p2'])

    def test_prepare_tasks(self):
        task = db_api.task_create(EXEC_ID, TASK.copy())
        tasks = [task]

        data_flow.prepare_tasks(tasks, CONTEXT)

        db_task = db_api.task_get(tasks[0]['id'])

        self.assertDictEqual(CONTEXT, db_task['in_context'])
        self.assertDictEqual({'p1': 'My string',
                              'p2': 'val32',
                              'p3': ''},
                             db_task['parameters'])

    def test_get_outbound_context(self):
        output = data_flow.get_task_output(TASK, {'new_key1': 'new_val1'})

        self.assertDictEqual(
            {
                'new_key11': 'new_val1',
                'task': {
                    'my_task': {
                        'new_key1': 'new_val1'
                    }
                }
            },
            output)



# Copyright 2011 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Simple class that stores security context information in the web request.

Projects should subclass this class if they wish to enhance the request
context or provide additional information in their specific WSGI pipeline.
"""

import itertools
import uuid


def generate_request_id():
    return b'req-' + str(uuid.uuid4()).encode('ascii')


class RequestContext(object):

    """Helper class to represent useful information about a request context.

    Stores information about the security context under which the user
    accesses the system, as well as additional request information.
    """

    user_idt_format = '{user} {tenant} {domain} {user_domain} {p_domain}'

    def __init__(self, auth_token=None, user=None, tenant=None, domain=None,
                 user_domain=None, project_domain=None, is_admin=False,
                 read_only=False, show_deleted=False, request_id=None,
                 instance_uuid=None):
        self.auth_token = auth_token
        self.user = user
        self.tenant = tenant
        self.domain = domain
        self.user_domain = user_domain
        self.project_domain = project_domain
        self.is_admin = is_admin
        self.read_only = read_only
        self.show_deleted = show_deleted
        self.instance_uuid = instance_uuid
        if not request_id:
            request_id = generate_request_id()
        self.request_id = request_id

    def to_dict(self):
        user_idt = (
            self.user_idt_format.format(user=self.user or '-',
                                        tenant=self.tenant or '-',
                                        domain=self.domain or '-',
                                        user_domain=self.user_domain or '-',
                                        p_domain=self.project_domain or '-'))

        return {'user': self.user,
                'tenant': self.tenant,
                'domain': self.domain,
                'user_domain': self.user_domain,
                'project_domain': self.project_domain,
                'is_admin': self.is_admin,
                'read_only': self.read_only,
                'show_deleted': self.show_deleted,
                'auth_token': self.auth_token,
                'request_id': self.request_id,
                'instance_uuid': self.instance_uuid,
                'user_identity': user_idt}


def get_admin_context(show_deleted=False):
    context = RequestContext(None,
                             tenant=None,
                             is_admin=True,
                             show_deleted=show_deleted)
    return context


def get_context_from_function_and_args(function, args, kwargs):
    """Find an arg of type RequestContext and return it.

       This is useful in a couple of decorators where we don't
       know much about the function we're wrapping.
    """

    for arg in itertools.chain(kwargs.values(), args):
        if isinstance(arg, RequestContext):
            return arg

    return None


def is_user_context(context):
    """Indicates if the request context is a normal user."""
    if not context:
        return False
    if context.is_admin:
        return False
    if not context.user_id or not context.project_id:
        return False
    return True



# -*- coding: utf-8 -*-
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import datetime
import uuid

from keystoneclient.middleware import auth_token
import mock
from oslo.config import cfg
import pecan
import pecan.testing

from mistral.db import api as db_api
from mistral.openstack.common import timeutils
from mistral.tests.api import base


WORKBOOKS = [
    {
        'name': "my_workbook",
        'description': "My cool Mistral workbook",
        'tags': ['deployment', 'demo']
    }
]


PKI_TOKEN_VERIFIED = {
    'token': {
        'methods': ['password'],
        'roles': [{'id': uuid.uuid4().hex,
                   'name': 'admin'}],
        'expires_at': timeutils.isotime(datetime.datetime.utcnow() +
                                        datetime.timedelta(seconds=60)),
        'project': {
            'domain': {'id': 'default', 'name': 'Default'},
            'id': uuid.uuid4().hex,
            'name': 'Mistral'
        },
        'catalog': [],
        'extras': {},
        'user': {
            'domain': {'id': 'default', 'name': 'Default'},
            'id': uuid.uuid4().hex,
            'name': 'admin'
        },
        'issued_at': timeutils.isotime()
    }
}


class TestKeystoneMiddleware(base.FunctionalTest):
    """Test that the keystone middleware AuthProtocol is executed
    when enabled.
    """

    def setUp(self):
        super(TestKeystoneMiddleware, self).setUp()
        cfg.CONF.set_default('auth_enable', True, group='pecan')
        self.app = pecan.testing.load_test_app({
            'app': {
                'root': cfg.CONF.pecan.root,
                'modules': cfg.CONF.pecan.modules,
                'debug': cfg.CONF.pecan.debug,
                'auth_enable': cfg.CONF.pecan.auth_enable
            }
        })

    @mock.patch.object(
        auth_token.AuthProtocol, '_get_user_token_from_header',
        mock.MagicMock(return_value=''))
    @mock.patch.object(
        auth_token.AuthProtocol, '_validate_user_token',
        mock.MagicMock(return_value=PKI_TOKEN_VERIFIED))
    @mock.patch.object(
        db_api, "workbook_get",
        mock.MagicMock(return_value=WORKBOOKS[0]))
    def test_auth_succeed(self):
        resp = self.app.get('/v1/workbooks/my_workbook')
        self.assertEqual(resp.status_int, 200)
        self.assertDictEqual(WORKBOOKS[0], resp.json)

    @mock.patch.object(
        auth_token.AuthProtocol, '_get_user_token_from_header',
        mock.MagicMock(return_value=''))
    @mock.patch.object(
        db_api, "workbook_get",
        mock.MagicMock(return_value=WORKBOOKS[0]))
    def test_auth_fail(self):
        # 401 unauthorized response is expected because the method
        # _validate_user_token is not mocked in this test.
        self.assertUnauthorized('/v1/workbooks/my_workbook')



# Copyright 2011 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import contextlib
import errno
import fcntl
import functools
import os
import shutil
import subprocess
import sys
import tempfile
import threading
import time
import weakref

from oslo.config import cfg

from mistral.openstack.common import fileutils
from mistral.openstack.common.gettextutils import _, _LE, _LI
from mistral.openstack.common import log as logging


LOG = logging.getLogger(__name__)


util_opts = [
    cfg.BoolOpt('disable_process_locking', default=False,
                help='Enables or disables inter-process locks.'),
    cfg.StrOpt('lock_path',
               default=os.environ.get("MISTRAL_LOCK_PATH"),
               help='Directory to use for lock files.')
]


CONF = cfg.CONF
CONF.register_opts(util_opts)


def set_defaults(lock_path):
    cfg.set_defaults(util_opts, lock_path=lock_path)


class _FileLock(object):
    """Lock implementation which allows multiple locks, working around
    issues like bugs.debian.org/cgi-bin/bugreport.cgi?bug=632857 and does
    not require any cleanup. Since the lock is always held on a file
    descriptor rather than outside of the process, the lock gets dropped
    automatically if the process crashes, even if __exit__ is not executed.

    There are no guarantees regarding usage by multiple green threads in a
    single process here. This lock works only between processes. Exclusive
    access between local threads should be achieved using the semaphores
    in the @synchronized decorator.

    Note these locks are released when the descriptor is closed, so it's not
    safe to close the file descriptor while another green thread holds the
    lock. Just opening and closing the lock file can break synchronisation,
    so lock files must be accessed only using this abstraction.
    """

    def __init__(self, name):
        self.lockfile = None
        self.fname = name

    def acquire(self):
        basedir = os.path.dirname(self.fname)

        if not os.path.exists(basedir):
            fileutils.ensure_tree(basedir)
            LOG.info(_LI('Created lock path: %s'), basedir)

        self.lockfile = open(self.fname, 'w')

        while True:
            try:
                # Using non-blocking locks since green threads are not
                # patched to deal with blocking locking calls.
                # Also upon reading the MSDN docs for locking(), it seems
                # to have a laughable 10 attempts "blocking" mechanism.
                self.trylock()
                LOG.debug('Got file lock "%s"', self.fname)
                return True
            except IOError as e:
                if e.errno in (errno.EACCES, errno.EAGAIN):
                    # external locks synchronise things like iptables
                    # updates - give it some time to prevent busy spinning
                    time.sleep(0.01)
                else:
                    raise threading.ThreadError(_("Unable to acquire lock on"
                                                  " `%(filename)s` due to"
                                                  " %(exception)s") %
                                                {
                                                    'filename': self.fname,
                                                    'exception': e,
                                                })

    def __enter__(self):
        self.acquire()
        return self

    def release(self):
        try:
            self.unlock()
            self.lockfile.close()
            LOG.debug('Released file lock "%s"', self.fname)
        except IOError:
            LOG.exception(_LE("Could not release the acquired lock `%s`"),
                          self.fname)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.release()

    def exists(self):
        return os.path.exists(self.fname)

    def trylock(self):
        raise NotImplementedError()

    def unlock(self):
        raise NotImplementedError()


class _WindowsLock(_FileLock):
    def trylock(self):
        msvcrt.locking(self.lockfile.fileno(), msvcrt.LK_NBLCK, 1)

    def unlock(self):
        msvcrt.locking(self.lockfile.fileno(), msvcrt.LK_UNLCK, 1)


class _FcntlLock(_FileLock):
    def trylock(self):
        fcntl.lockf(self.lockfile, fcntl.LOCK_EX | fcntl.LOCK_NB)

    def unlock(self):
        fcntl.lockf(self.lockfile, fcntl.LOCK_UN)


class _PosixLock(object):
    def __init__(self, name):
        # Hash the name because it's not valid to have POSIX semaphore
        # names with things like / in them. Then use base64 to encode
        # the digest() instead taking the hexdigest() because the
        # result is shorter and most systems can't have shm sempahore
        # names longer than 31 characters.
        h = hashlib.sha1()
        h.update(name.encode('ascii'))
        self.name = str((b'/' + base64.urlsafe_b64encode(
            h.digest())).decode('ascii'))

    def acquire(self, timeout=None):
        self.semaphore = posix_ipc.Semaphore(self.name,
                                             flags=posix_ipc.O_CREAT,
                                             initial_value=1)
        self.semaphore.acquire(timeout)
        return self

    def __enter__(self):
        self.acquire()
        return self

    def release(self):
        self.semaphore.release()
        self.semaphore.close()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.release()

    def exists(self):
        try:
            semaphore = posix_ipc.Semaphore(self.name)
        except posix_ipc.ExistentialError:
            return False
        else:
            semaphore.close()
        return True


if os.name == 'nt':
    import msvcrt
    InterProcessLock = _WindowsLock
    FileLock = _WindowsLock
else:
    import base64
    import hashlib
    import posix_ipc
    InterProcessLock = _PosixLock
    FileLock = _FcntlLock

_semaphores = weakref.WeakValueDictionary()
_semaphores_lock = threading.Lock()


def _get_lock_path(name, lock_file_prefix, lock_path=None):
    # NOTE(mikal): the lock name cannot contain directory
    # separators
    name = name.replace(os.sep, '_')
    if lock_file_prefix:
        sep = '' if lock_file_prefix.endswith('-') else '-'
        name = '%s%s%s' % (lock_file_prefix, sep, name)

    local_lock_path = lock_path or CONF.lock_path

    if not local_lock_path:
        # NOTE(bnemec): Create a fake lock path for posix locks so we don't
        # unnecessarily raise the RequiredOptError below.
        if InterProcessLock is not _PosixLock:
            raise cfg.RequiredOptError('lock_path')
        local_lock_path = 'posixlock:/'

    return os.path.join(local_lock_path, name)


def external_lock(name, lock_file_prefix=None, lock_path=None):
    LOG.debug('Attempting to grab external lock "%(lock)s"',
              {'lock': name})

    lock_file_path = _get_lock_path(name, lock_file_prefix, lock_path)

    # NOTE(bnemec): If an explicit lock_path was passed to us then it
    # means the caller is relying on file-based locking behavior, so
    # we can't use posix locks for those calls.
    if lock_path:
        return FileLock(lock_file_path)
    return InterProcessLock(lock_file_path)


def remove_external_lock_file(name, lock_file_prefix=None):
    """Remove a external lock file when it's not used anymore
    This will be helpful when we have a lot of lock files
    """
    with internal_lock(name):
        lock_file_path = _get_lock_path(name, lock_file_prefix)
        try:
            os.remove(lock_file_path)
        except OSError:
            LOG.info(_LI('Failed to remove file %(file)s'),
                     {'file': lock_file_path})


def internal_lock(name):
    with _semaphores_lock:
        try:
            sem = _semaphores[name]
        except KeyError:
            sem = threading.Semaphore()
            _semaphores[name] = sem

    LOG.debug('Got semaphore "%(lock)s"', {'lock': name})
    return sem


@contextlib.contextmanager
def lock(name, lock_file_prefix=None, external=False, lock_path=None):
    """Context based lock

    This function yields a `threading.Semaphore` instance (if we don't use
    eventlet.monkey_patch(), else `semaphore.Semaphore`) unless external is
    True, in which case, it'll yield an InterProcessLock instance.

    :param lock_file_prefix: The lock_file_prefix argument is used to provide
      lock files on disk with a meaningful prefix.

    :param external: The external keyword argument denotes whether this lock
      should work across multiple processes. This means that if two different
      workers both run a method decorated with @synchronized('mylock',
      external=True), only one of them will execute at a time.
    """
    int_lock = internal_lock(name)
    with int_lock:
        if external and not CONF.disable_process_locking:
            ext_lock = external_lock(name, lock_file_prefix, lock_path)
            with ext_lock:
                yield ext_lock
        else:
            yield int_lock
    LOG.debug('Released semaphore "%(lock)s"', {'lock': name})


def synchronized(name, lock_file_prefix=None, external=False, lock_path=None):
    """Synchronization decorator.

    Decorating a method like so::

        @synchronized('mylock')
        def foo(self, *args):
           ...

    ensures that only one thread will execute the foo method at a time.

    Different methods can share the same lock::

        @synchronized('mylock')
        def foo(self, *args):
           ...

        @synchronized('mylock')
        def bar(self, *args):
           ...

    This way only one of either foo or bar can be executing at a time.
    """

    def wrap(f):
        @functools.wraps(f)
        def inner(*args, **kwargs):
            try:
                with lock(name, lock_file_prefix, external, lock_path):
                    LOG.debug('Got semaphore / lock "%(function)s"',
                              {'function': f.__name__})
                    return f(*args, **kwargs)
            finally:
                LOG.debug('Semaphore / lock released "%(function)s"',
                          {'function': f.__name__})
        return inner
    return wrap


def synchronized_with_prefix(lock_file_prefix):
    """Partial object generator for the synchronization decorator.

    Redefine @synchronized in each project like so::

        (in nova/utils.py)
        from nova.openstack.common import lockutils

        synchronized = lockutils.synchronized_with_prefix('nova-')


        (in nova/foo.py)
        from nova import utils

        @utils.synchronized('mylock')
        def bar(self, *args):
           ...

    The lock_file_prefix argument is used to provide lock files on disk with a
    meaningful prefix.
    """

    return functools.partial(synchronized, lock_file_prefix=lock_file_prefix)


def main(argv):
    """Create a dir for locks and pass it to command from arguments

    If you run this:
    python -m openstack.common.lockutils python setup.py testr <etc>

    a temporary directory will be created for all your locks and passed to all
    your tests in an environment variable. The temporary dir will be deleted
    afterwards and the return value will be preserved.
    """

    lock_dir = tempfile.mkdtemp()
    os.environ["MISTRAL_LOCK_PATH"] = lock_dir
    try:
        ret_val = subprocess.call(argv[1:])
    finally:
        shutil.rmtree(lock_dir, ignore_errors=True)
    return ret_val


if __name__ == '__main__':
    sys.exit(main(sys.argv))



# -*- coding: utf-8 -*-
#
# Copyright 2013 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import six

from mistral.workbook import base


class TaskSpec(base.BaseSpec):
    _required_keys = ['name', 'action']

    def __init__(self, task):
        super(TaskSpec, self).__init__(task)

        self._prepare(task)

        if self.validate():
            self.requires = task['requires']
            self.action = task['action']
            self.name = task['name']
            self.parameters = task.get('parameters', {})

    def _prepare(self, task):
        if task:
            req = task.get("requires", {})

            if req and isinstance(req, list):
                task["requires"] = dict(zip(req, [''] * len(req)))
            elif isinstance(req, dict):
                task['requires'] = req

    def _get_as_dict(self, key):
        tasks = self.get_property(key)

        if not tasks:
            return {}

        if isinstance(tasks, dict):
            return tasks
        elif isinstance(tasks, list):
            result = {}
            for t in tasks:
                result.update(t if isinstance(t, dict) else {t: ''})
            return result
        elif isinstance(tasks, six.string_types):
            return {tasks: ''}

    def get_property(self, property_name, default=None):
        return self._data.get(property_name, default)

    def get_requires(self):
        return self._get_as_dict('requires').keys()

    def get_on_error(self):
        return self._get_as_dict("on-error")

    def get_on_success(self):
        return self._get_as_dict("on-success")

    def get_on_finish(self):
        return self._get_as_dict("on-finish")

    def get_action_namespace(self):
        return self.action.split('.')[0]

    def get_action_name(self):
        return self.action.split('.')[1]

    def get_full_action_name(self):
        return self.action

    def is_retry_task(self):
        return self.get_property("retry") is not None

    def get_retry_parameters(self):
        iterations = 0
        break_on = None
        delay = 0
        retry = self.get_property("retry")

        if retry:
            if "count" in retry:
                iterations = retry["count"]
            if "break-on" in retry:
                break_on = retry["break-on"]
            if "delay" in retry:
                delay = retry["delay"]

        return iterations, break_on, delay


class TaskSpecList(base.BaseSpecList):
    item_class = TaskSpec



# -*- coding: utf-8 -*-
#
# Copyright 2013 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import json
import mock

from mistral.db import api as db_api
from mistral import engine
from mistral.tests.api import base

# TODO(everyone): later we need additional tests verifying all the errors etc.

TASKS = [
    {
        'id': "1",
        'workbook_name': "my_workbook",
        'execution_id': '123',
        'name': 'my_task',
        'description': 'My cool task',
        'state': 'RUNNING',
        'tags': ['deployment', 'demo'],
        'output': {
            'a': 'b'
        },
        'parameters': {
            'c': 'd'
        }
    }
]

UPDATED_TASK = TASKS[0].copy()
UPDATED_TASK['state'] = 'STOPPED'


def canonize(json_dict):
    if json_dict.get('output'):
        json_dict['output'] = json.loads(json_dict['output'])

    if json_dict.get('parameters'):
        json_dict['parameters'] = json.loads(json_dict['parameters'])

    return json_dict


class TestTasksController(base.FunctionalTest):
    @mock.patch.object(db_api, "task_get",
                       mock.MagicMock(return_value=TASKS[0]))
    def test_workbook_get(self):
        resp = self.app.get('/v1/workbooks/my_workbook/executions/123/tasks/1')

        self.assertEqual(resp.status_int, 200)
        self.assertDictEqual(TASKS[0], canonize(resp.json))

    @mock.patch.object(db_api, "task_get",
                       mock.MagicMock(return_value=TASKS[0]))
    def test_execution_get(self):
        resp = self.app.get('/v1/executions/123/tasks/1')

        self.assertEqual(resp.status_int, 200)
        self.assertDictEqual(TASKS[0], canonize(resp.json))

    @mock.patch.object(db_api, "task_get",
                       mock.MagicMock(return_value=TASKS[0]))
    def test_root_get(self):
        resp = self.app.get('/v1/tasks/1')

        self.assertEqual(resp.status_int, 200)
        self.assertDictEqual(TASKS[0], canonize(resp.json))

    @mock.patch.object(engine.EngineClient, "convey_task_result",
                       mock.MagicMock(return_value=UPDATED_TASK))
    @mock.patch.object(db_api, "task_get",
                       mock.MagicMock(return_value=TASKS[0]))
    def test_workbook_put(self):
        resp = self.app.put_json(
            '/v1/workbooks/my_workbook/executions/123/tasks/1',
            dict(state='STOPPED'))
        self.assertEqual(resp.status_int, 200)
        self.assertDictEqual(UPDATED_TASK, canonize(resp.json))

    @mock.patch.object(engine.EngineClient, "convey_task_result",
                       mock.MagicMock(return_value=UPDATED_TASK))
    @mock.patch.object(db_api, "task_get",
                       mock.MagicMock(return_value=TASKS[0]))
    def test_execution_put(self):
        resp = self.app.put_json(
            '/v1/executions/123/tasks/1',
            dict(state='STOPPED'))
        self.assertEqual(resp.status_int, 200)
        self.assertDictEqual(UPDATED_TASK, canonize(resp.json))

    @mock.patch.object(engine.EngineClient, "convey_task_result",
                       mock.MagicMock(return_value=UPDATED_TASK))
    @mock.patch.object(db_api, "task_get",
                       mock.MagicMock(return_value=TASKS[0]))
    def test_root_put(self):
        resp = self.app.put_json(
            '/v1/tasks/1',
            dict(state='STOPPED'))
        self.assertEqual(resp.status_int, 200)
        self.assertDictEqual(UPDATED_TASK, canonize(resp.json))

    @mock.patch.object(engine.EngineClient, "convey_task_result",
                       mock.MagicMock(return_value=UPDATED_TASK))
    def test_put_no_task(self):
        resp = self.app.put_json(
            '/v1/workbooks/my_workbook/executions/123/tasks/1',
            dict(state='STOPPED'), expect_errors=True)
        self.assertEqual(resp.status_int, 404)

    @mock.patch.object(db_api, "tasks_get",
                       mock.MagicMock(return_value=TASKS))
    @mock.patch.object(db_api, "ensure_execution_exists",
                       mock.MagicMock(return_value={'id': "abc123"}))
    def test_workbook_get_all(self):
        resp = self.app.get('/v1/workbooks/my_workbook/executions/123/tasks')

        self.assertEqual(resp.status_int, 200)

        self.assertEqual(len(resp.json), 1)
        self.assertDictEqual(TASKS[0], canonize(resp.json['tasks'][0]))

    @mock.patch.object(db_api, "tasks_get",
                       mock.MagicMock(return_value=TASKS))
    @mock.patch.object(db_api, "ensure_execution_exists",
                       mock.MagicMock(return_value={'id': "abc123"}))
    def test_execution_get_all(self):
        resp = self.app.get('/v1/executions/123/tasks')

        self.assertEqual(resp.status_int, 200)

        self.assertEqual(len(resp.json), 1)
        self.assertDictEqual(TASKS[0], canonize(resp.json['tasks'][0]))

    @mock.patch.object(db_api, "tasks_get",
                       mock.MagicMock(return_value=TASKS))
    @mock.patch.object(db_api, "ensure_execution_exists",
                       mock.MagicMock(return_value={'id': "abc123"}))
    def test_root_get_all(self):
        resp = self.app.get('/v1/tasks')

        self.assertEqual(resp.status_int, 200)

        self.assertEqual(len(resp.json), 1)
        self.assertDictEqual(TASKS[0], canonize(resp.json['tasks'][0]))

    @mock.patch.object(db_api, "tasks_get",
                       mock.MagicMock(return_value=TASKS))
    def test_get_all_nonexistent_execution(self):
        self.assertNotFound('/v1/workbooks/my_workbook/executions/123/tasks')



# -*- coding: utf-8 -*-
#
# Copyright 2013 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from mistral import dsl_parser as parser
from mistral.tests import base


SIMPLE_WORKBOOK = """
Workflow:
  tasks:
    create-vms:
      action: MyRest.create-vm
"""


class DSLModelTest(base.BaseTest):
    def setUp(self):
        super(DSLModelTest, self).setUp()
        self.doc = base.get_resource("test_rest.yaml")

    def test_load_dsl(self):
        wb = parser.get_workbook(self.doc)

        self.assertEqual(wb.workflow.tasks.items, wb.tasks.items)
        self.assertEqual(wb.tasks.get("create-vms").name, "create-vms")
        self.assertEqual(4, len(wb.namespaces.get("MyRest").actions))

    def test_tasks(self):
        wb = parser.get_workbook(self.doc)
        self.assertEqual(len(wb.tasks), 6)

        attach_volumes = wb.tasks.get("attach-volumes")

        self.assertEqual(attach_volumes.get_action_namespace(), "MyRest")

        t_parameters = {"image_id": 1234, "flavor_id": 2}
        create_vm_nova = wb.tasks.get("create-vm-nova")

        self.assertEqual(create_vm_nova.parameters, t_parameters)

        attach_requires = {"create-vms": ''}

        self.assertEqual(attach_volumes.requires, attach_requires)

        subsequent = wb.tasks.get("test_subsequent")

        subseq_success = subsequent.get_on_success()
        subseq_error = subsequent.get_on_error()
        subseq_finish = subsequent.get_on_finish()

        self.assertEqual(subseq_success, {"attach-volumes": ''})
        self.assertEqual(subseq_error, {"backup-vms": "$.status != 'OK'",
                                        "attach-volumes": ''})
        self.assertEqual(subseq_finish, {"create-vms": ''})

    def test_actions(self):
        wb = parser.get_workbook(self.doc)

        actions = wb.namespaces.get("MyRest").actions

        self.assertEqual(len(actions), 4)

        create_vm = actions.get("create-vm")

        base_params = create_vm.base_parameters

        self.assertEqual('std.mistral_http', create_vm.clazz)
        self.assertIn('method', base_params)
        self.assertIn('headers', base_params)
        self.assertEqual('$.auth_token',
                         base_params['headers']['X-Auth-Token'])
        self.assertEqual('application/json',
                         base_params['headers']['Content-Type'])

        action = wb.get_action("MyRest.create-vm")

        self.assertIsNotNone(action)
        self.assertEqual("create-vm", action.name)
        self.assertEqual("MyRest", action.namespace)

    def test_namespaces(self):
        wb = parser.get_workbook(self.doc)

        self.assertEqual(len(wb.namespaces), 2)

        nova_namespace = wb.namespaces.get("Nova")

        self.assertEqual(1, len(nova_namespace.actions))

    def test_workbook_without_namespaces(self):
        parser.get_workbook(SIMPLE_WORKBOOK)

    def test_triggers(self):
        wb = parser.get_workbook(self.doc)

        self.assertEqual(len(wb.get_triggers()), 1)



# -*- coding: utf-8 -*-
#
# Copyright 2014 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import json

import mock
import requests

from mistral.actions import std_actions as std
from mistral.tests import base


URL = "http://some_url"

DATA = {
    "server": {
        "id": "12345",
        "metadata": {
            "name": "super_server"
        }
    }
}


def get_fake_response():
    response = mock.Mock()

    response.content = json.dumps(DATA)
    response.json = mock.MagicMock(return_value=DATA)
    response.headers = {'Content-Type': 'application/json'}
    response.status_code = 200

    return response


class HTTPActionTest(base.BaseTest):
    @mock.patch.object(requests, "request")
    def test_http_action(self, mocked_method):
        mocked_method.return_value = get_fake_response()

        action = std.HTTPAction(
            url=URL,
            method="POST",
            body=DATA,
            timeout=20,
            allow_redirects=True)

        DATA_STR = json.dumps(DATA)

        self.assertEqual(DATA_STR, action.body)
        self.assertEqual(URL, action.url)

        result = action.run()

        self.assertIsInstance(result, dict)
        self.assertEqual(DATA, result['content'])
        self.assertIn('headers', result)
        self.assertEqual(200, result['status'])

        mocked_method.assert_called_with(
            "POST",
            URL,
            data=DATA_STR,
            headers=None,
            cookies=None,
            params=None,
            timeout=20,
            auth=None,
            allow_redirects=True,
            proxies=None
        )



# -*- coding: utf-8 -*-
#
# Copyright 2013 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from mistral import context
from mistral.db import api as db_api
from mistral import dsl_parser as parser
from mistral import engine
from mistral.openstack.common import log
from mistral.openstack.common import periodic_task
from mistral.openstack.common import threadgroup
from mistral.services import scheduler as sched
from mistral.services import trusts


LOG = log.getLogger(__name__)


class MistralPeriodicTasks(periodic_task.PeriodicTasks):

    def __init__(self, transport=None):
        super(MistralPeriodicTasks, self).__init__()
        self.transport = engine.get_transport(transport)
        self.engine = engine.EngineClient(self.transport)

    @periodic_task.periodic_task(spacing=1, run_immediately=True)
    def scheduler_triggers(self, ctx):
        LOG.debug('Processing next Scheduler triggers.')

        for trigger in sched.get_next_triggers():
            wb = db_api.workbook_get(trigger['workbook_name'])
            context.set_ctx(trusts.create_context(wb))

            try:
                task = parser.get_workbook(
                    wb['definition']).get_trigger_task_name(trigger['name'])
                self.engine.start_workflow_execution(wb['name'], task)
            finally:
                sched.set_next_execution_time(trigger)
                context.set_ctx(None)


def setup(transport):
    tg = threadgroup.ThreadGroup()
    pt = MistralPeriodicTasks(transport=transport)

    tg.add_dynamic_timer(
        pt.run_periodic_tasks,
        initial_delay=None,
        periodic_interval_max=1,
        context=None)



# -*- coding: utf-8 -*-
#
# Copyright 2013 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import datetime

from mistral.openstack.common import timeutils
from mistral.services import scheduler as s
from mistral.tests import base


SAMPLE_TRIGGER = {
    "id": "123",
    "name": "test_trigger",
    "patter": "* *",
    "next_execution_time": timeutils.utcnow(),
    "workbook_name": "My workbook"
}


class SchedulerTest(base.DbTestCase):
    def setUp(self):
        super(SchedulerTest, self).setUp()
        self.wb_name = "My workbook"

    def test_trigger_create_and_update(self):
        base = datetime.datetime(2010, 8, 25)
        next_trigger = datetime.datetime(2010, 8, 25, 0, 5)
        trigger = s.create_trigger("test", "*/5 * * * *", self.wb_name, base)
        self.assertEqual(trigger['next_execution_time'], next_trigger)

        trigger = s.set_next_execution_time(trigger)
        next_trigger = datetime.datetime(2010, 8, 25, 0, 10)
        self.assertEqual(trigger['next_execution_time'], next_trigger)

    def test_get_trigger_in_correct_orders(self):
        base = datetime.datetime(2010, 8, 25)
        s.create_trigger("test1", "*/5 * * * *", self.wb_name, base)
        base = datetime.datetime(2010, 8, 22)
        s.create_trigger("test2", "*/5 * * * *", self.wb_name, base)
        base = datetime.datetime(2010, 9, 21)
        s.create_trigger("test3", "*/5 * * * *", self.wb_name, base)
        base = datetime.datetime.now() + datetime.timedelta(0, 50)
        s.create_trigger("test4", "*/5 * * * *", self.wb_name, base)
        triggersName = [e['name'] for e in s.get_next_triggers()]

        self.assertEqual(triggersName, ["test2", "test1", "test3"])



# -*- coding: utf-8 -*-
#
# Copyright 2013 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import json

import pecan
from pecan import rest
from wsme import types as wtypes
import wsmeext.pecan as wsme_pecan

from mistral.api.controllers import resource
from mistral.db import api as db_api
from mistral.openstack.common import log as logging
from mistral.utils import rest_utils


LOG = logging.getLogger(__name__)


class Task(resource.Resource):
    """Task resource."""

    id = wtypes.text
    "The ID of the Task."

    workbook_name = wtypes.text
    execution_id = wtypes.text
    name = wtypes.text
    description = wtypes.text
    state = wtypes.text
    tags = [wtypes.text]
    output = wtypes.text
    parameters = wtypes.text

    @classmethod
    def from_dict(cls, d):
        e = cls()

        for key, val in d.items():
            if hasattr(e, key):
                # Nonetype check for dictionary must be explicit
                if val is not None and (
                        key == 'parameters' or key == 'output'):
                    val = json.dumps(val)
                setattr(e, key, val)

        return e

    @classmethod
    def sample(cls):
        return cls(id='1234',
                   workbook_name='notifier',
                   execution_id='234',
                   name='build_greeting',
                   description='tell when you are done',
                   state='OK',
                   tags=['foo', 'fee'],
                   output='{"task": {"build_greeting": '
                          '{"greeting": "Hello, John Doe!"}}}',
                   parameters='{ "first_name": "John", "last_name": "Doe"}')


class Tasks(resource.Resource):
    """A collection of tasks."""

    tasks = [Task]
    "List of tasks."

    @classmethod
    def sample(cls):
        return cls(tasks=[Task.sample()])


class TasksController(rest.RestController):
    def _get(self, id):
        values = db_api.task_get(id)
        return Task.from_dict(values)

    def _put(self, id, task):
        if db_api.task_get(id):
            # TODO(rakhmerov): pass task result once it's implemented
            engine = pecan.request.context['engine']
            values = engine.convey_task_result(id,
                                               task.state, None)

            return Task.from_dict(values)

    def _get_all(self, **kwargs):
        tasks = [Task.from_dict(values)
                 for values in db_api.tasks_get(**kwargs)]

        return Tasks(tasks=tasks)


class WorkbookTasksController(TasksController):
    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(Task, wtypes.text, wtypes.text, wtypes.text)
    def get(self, workbook_name, execution_id, id):
        """Return the specified task."""
        LOG.debug("Fetch task [workbook_name=%s, execution_id=%s, id=%s]" %
                  (workbook_name, execution_id, id))

        return self._get(id)

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(Task, wtypes.text, wtypes.text, wtypes.text,
                         body=Task)
    def put(self, workbook_name, execution_id, id, task):
        """Update the specified task."""
        LOG.debug("Update task "
                  "[workbook_name=%s, execution_id=%s, id=%s, task=%s]" %
                  (workbook_name, execution_id, id, task))

        return self._put(id, task)

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(Tasks, wtypes.text, wtypes.text)
    def get_all(self, workbook_name, execution_id):
        """Return all tasks within the execution."""
        db_api.ensure_execution_exists(execution_id)

        LOG.debug("Fetch tasks [workbook_name=%s, execution_id=%s]" %
                  (workbook_name, execution_id))

        return self._get_all(workbook_name=workbook_name,
                             execution_id=execution_id)


class ExecutionTasksController(TasksController):
    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(Task, wtypes.text, wtypes.text)
    def get(self, execution_id, id):
        """Return the specified task."""
        LOG.debug("Fetch task [execution_id=%s, id=%s]" %
                  (execution_id, id))

        return self._get(id)

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(Task, wtypes.text, wtypes.text,
                         body=Task)
    def put(self, execution_id, id, task):
        """Update the specified task."""
        LOG.debug("Update task "
                  "[execution_id=%s, id=%s, task=%s]" %
                  (execution_id, id, task))

        return self._put(id, task)

    @wsme_pecan.wsexpose(Tasks, wtypes.text)
    def get_all(self, execution_id):
        """Return all tasks within the execution."""
        LOG.debug("Fetch tasks [execution_id=%s]" % execution_id)

        return self._get_all(execution_id=execution_id)


class RootTasksController(TasksController):
    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(Task, wtypes.text)
    def get(self, id):
        """Return the specified task."""
        LOG.debug("Fetch task [id=%s]" % id)

        return self._get(id)

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(Task, wtypes.text,
                         body=Task)
    def put(self, id, task):
        """Update the specified task."""
        LOG.debug("Update task "
                  "[id=%s, task=%s]" %
                  (id, task))

        return self._put(id, task)

    @wsme_pecan.wsexpose(Tasks)
    def get_all(self):
        """Return all tasks within the execution."""
        LOG.debug("Fetch tasks")

        return self._get_all()



# -*- coding: utf-8 -*-
#
# Copyright 2013 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
Configuration options registration and useful routines.
"""

from oslo.config import cfg

from mistral.openstack.common import log
from mistral import version

api_opts = [
    cfg.StrOpt('host', default='0.0.0.0', help='Mistral API server host'),
    cfg.IntOpt('port', default=8989, help='Mistral API server port')
]

engine_opts = [
    cfg.StrOpt('engine', default='default',
               help='Mistral engine plugin'),
    cfg.StrOpt('host', default='0.0.0.0',
               help='Name of the engine node. This can be an opaque '
                    'identifier. It is not necessarily a hostname, '
                    'FQDN, or IP address.'),
    cfg.StrOpt('topic', default='engine',
               help='The message topic that the engine listens on.'),
    cfg.StrOpt('version', default='1.0',
               help='The version of the engine.')
]

pecan_opts = [
    cfg.StrOpt('root', default='mistral.api.controllers.root.RootController',
               help='Pecan root controller'),
    cfg.ListOpt('modules', default=["mistral.api"],
                help='A list of modules where pecan will search for '
                     'applications.'),
    cfg.BoolOpt('debug', default=False,
                help='Enables the ability to display tracebacks in the '
                     'browser and interactively debug during '
                     'development.'),
    cfg.BoolOpt('auth_enable', default=True,
                help='Enables user authentication in pecan.')
]

db_opts = [
    # TODO(everyone): add DB properties.
]

use_debugger = cfg.BoolOpt(
    "use-debugger",
    default=False,
    help='Enables debugger. Note that using this option changes how the '
    'eventlet library is used to support async IO. This could result '
    'in failures that do not occur under normal operation. '
    'Use at your own risk.'
)

executor_opts = [
    cfg.StrOpt('host', default='0.0.0.0',
               help='Name of the executor node. This can be an opaque '
                    'identifier. It is not necessarily a hostname, '
                    'FQDN, or IP address.'),
    cfg.StrOpt('topic', default='executor',
               help='The message topic that the executor listens on.'),
    cfg.StrOpt('version', default='1.0',
               help='The version of the executor.')
]

launch_opt = cfg.ListOpt(
    'server',
    default=['all'],
    help='Specifies which mistral server to start by the launch script. '
         'Valid options are all or any combination of '
         'api, engine, and executor.'
)

action_plugins_opt = cfg.ListOpt(
    'action_plugins',
    default=['mistral.actions.std'],
    help='List of namespaces to search for plug-ins.')


wf_trace_log_name_opt = cfg.StrOpt('workflow_trace_log_name',
                                   default='workflow_trace',
                                   help='Logger name for pretty '
                                   'workflow trace output.')

CONF = cfg.CONF

CONF.register_opts(api_opts, group='api')
CONF.register_opts(engine_opts, group='engine')
CONF.register_opts(pecan_opts, group='pecan')
CONF.register_opts(db_opts, group='database')
CONF.register_opts(executor_opts, group='executor')
CONF.register_opt(wf_trace_log_name_opt)
CONF.register_opt(action_plugins_opt)

CONF.register_cli_opt(use_debugger)
CONF.register_cli_opt(launch_opt)

CONF.import_opt('verbose', 'mistral.openstack.common.log')
CONF.import_opt('debug', 'mistral.openstack.common.log')
CONF.import_opt('log_dir', 'mistral.openstack.common.log')
CONF.import_opt('log_file', 'mistral.openstack.common.log')
CONF.import_opt('log_config_append', 'mistral.openstack.common.log')
CONF.import_opt('log_format', 'mistral.openstack.common.log')
CONF.import_opt('log_date_format', 'mistral.openstack.common.log')
CONF.import_opt('use_syslog', 'mistral.openstack.common.log')
CONF.import_opt('syslog_log_facility', 'mistral.openstack.common.log')

# Extend oslo default_log_levels to include some that are useful for mistral
# some are in oslo logging already, this is just making sure it stays this
# way.
default_log_levels = cfg.CONF.default_log_levels
logs_to_quieten = ['sqlalchemy=WARN',
                   'oslo.messaging=INFO',
                   'iso8601=WARN',
                   'eventlet.wsgi.server=WARN',
                   'stevedore=INFO',
                   'mistral.openstack.common.loopingcall=INFO',
                   'mistral.openstack.common.periodic_task=INFO',
                   'mistral.services.periodic=INFO']
for chatty in logs_to_quieten:
    if chatty not in default_log_levels:
        default_log_levels.append(chatty)
cfg.set_defaults(log.log_opts,
                 default_log_levels=default_log_levels)


def parse_args(args=None, usage=None, default_config_files=None):
    CONF(args=args,
         project='mistral',
         version=version,
         usage=usage,
         default_config_files=default_config_files)



# -*- coding: utf-8 -*-
#
# Copyright 2013 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from mistral import context as auth_context
from mistral.db.sqlalchemy import api as db_api
from mistral import exceptions as exc
from mistral.openstack.common import timeutils
from mistral.tests import base as test_base


TRIGGERS = [
    {
        'id': '1',
        'name': 'test_trigger1',
        'workbook_name': 'my_workbook1',
        'pattern': '* *',
        'next_execution_time': timeutils.utcnow(),
        'updated_at': None
    },
    {
        'id': '2',
        'name': 'test_trigger2',
        'workbook_name': 'my_workbook2',
        'pattern': '* * *',
        'next_execution_time': timeutils.utcnow(),
        'updated_at': None
    }
]


class TriggerTest(test_base.DbTestCase):
    def test_trigger_create_and_get(self):
        created = db_api.trigger_create(TRIGGERS[0])
        self.assertIsInstance(created, dict)

        fetched = db_api.trigger_get(created['id'])
        self.assertIsInstance(fetched, dict)
        self.assertDictEqual(created, fetched)

    def test_trigger_update(self):
        created = db_api.trigger_create(TRIGGERS[0])
        self.assertIsInstance(created, dict)

        updated = db_api.trigger_update(created['id'], {'pattern': '0 * *'})
        self.assertIsInstance(updated, dict)
        self.assertEqual('0 * *', updated['pattern'])

        fetched = db_api.trigger_get(created['id'])
        self.assertDictEqual(updated, fetched)

    def test_trigger_delete(self):
        created = db_api.trigger_create(TRIGGERS[0])
        db_api.trigger_delete(created['id'])

        self.assertRaises(exc.NotFoundException, db_api.trigger_get,
                          created['id'])

    def test_trigger_list(self):
        created0 = db_api.trigger_create(TRIGGERS[0])
        created1 = db_api.trigger_create(TRIGGERS[1])

        fetched = db_api.triggers_get_all()

        self.assertEqual(2, len(fetched))
        self.assertDictEqual(created0, fetched[0])
        self.assertDictEqual(created1, fetched[1])


WORKBOOKS = [
    {
        'id': '1',
        'name': 'my_workbook1',
        'description': 'my description',
        'definition': 'empty',
        'tags': ['mc'],
        'scope': 'public',
        'updated_at': None,
        'project_id': '1233',
        'trust_id': '1234'
    },
    {
        'id': '2',
        'name': 'my_workbook2',
        'description': 'my description',
        'definition': 'empty',
        'tags': ['mc'],
        'scope': 'private',
        'updated_at': None,
        'project_id': '1233',
        'trust_id': '12345'
    },
]


class WorkbookTest(test_base.DbTestCase):
    def test_workbook_create_and_get(self):
        created = db_api.workbook_create(WORKBOOKS[0])
        self.assertIsInstance(created, dict)

        fetched = db_api.workbook_get(created['name'])
        self.assertIsInstance(fetched, dict)
        self.assertDictEqual(created, fetched)

    def test_workbook_update(self):
        created = db_api.workbook_create(WORKBOOKS[0])
        self.assertIsInstance(created, dict)

        updated = db_api.workbook_update(created['name'],
                                         {'description': 'my new desc'})
        self.assertIsInstance(updated, dict)
        self.assertEqual('my new desc', updated['description'])

        fetched = db_api.workbook_get(created['name'])
        self.assertDictEqual(updated, fetched)

    def test_workbook_list(self):
        created0 = db_api.workbook_create(WORKBOOKS[0])
        created1 = db_api.workbook_create(WORKBOOKS[1])

        fetched = db_api.workbooks_get_all()

        self.assertEqual(2, len(fetched))
        self.assertDictEqual(created0, fetched[0])
        self.assertDictEqual(created1, fetched[1])

    def test_workbook_delete(self):
        created = db_api.workbook_create(WORKBOOKS[0])
        self.assertIsInstance(created, dict)

        fetched = db_api.workbook_get(created['name'])
        self.assertIsInstance(fetched, dict)
        self.assertDictEqual(created, fetched)

        db_api.workbook_delete(created['name'])
        self.assertRaises(exc.NotFoundException,
                          db_api.workbook_get, created['name'])

    def test_workbook_private(self):
        # create a workbook(scope=private) as under one project
        # then make sure it's NOT visible for other projects.
        created1 = db_api.workbook_create(WORKBOOKS[1])

        fetched = db_api.workbooks_get_all()

        self.assertEqual(1, len(fetched))
        self.assertDictEqual(created1, fetched[0])

        # create a new user.
        ctx = auth_context.MistralContext(user_id='9-0-44-5',
                                          project_id='99-88-33',
                                          user_name='test-user',
                                          project_name='test-another',
                                          is_admin=False)
        auth_context.set_ctx(ctx)

        fetched = db_api.workbooks_get_all()
        self.assertEqual(0, len(fetched))

    def test_workbook_public(self):
        # create a workbook(scope=public) as under one project
        # then make sure it's visible for other projects.
        created0 = db_api.workbook_create(WORKBOOKS[0])

        fetched = db_api.workbooks_get_all()

        self.assertEqual(1, len(fetched))
        self.assertDictEqual(created0, fetched[0])

        # assert that the project_id stored is actually the context's
        # project_id not the one given.
        self.assertEqual(created0['project_id'], auth_context.ctx().project_id)
        self.assertNotEqual(WORKBOOKS[0]['project_id'],
                            auth_context.ctx().project_id)

        # create a new user.
        ctx = auth_context.MistralContext(user_id='9-0-44-5',
                                          project_id='99-88-33',
                                          user_name='test-user',
                                          project_name='test-another',
                                          is_admin=False)
        auth_context.set_ctx(ctx)

        fetched = db_api.workbooks_get_all()
        self.assertEqual(1, len(fetched))
        self.assertDictEqual(created0, fetched[0])
        self.assertEqual('public', created0['scope'])


EXECUTIONS = [
    {
        'id': '1',
        'workbook_name': 'my_workbook',
        'task': 'my_task1',
        'state': 'IDLE',
        'updated_at': None,
        'context': None
    },
    {
        'id': '2',
        'workbook_name': 'my_workbook',
        'task': 'my_task2',
        'state': 'RUNNING',
        'updated_at': None,
        'context': {'image_id': '123123'}
    }
]


class ExecutionTest(test_base.DbTestCase):
    def test_execution_create_and_get(self):
        created = db_api.execution_create(EXECUTIONS[0]['workbook_name'],
                                          EXECUTIONS[0])
        self.assertIsInstance(created, dict)

        fetched = db_api.execution_get(created['id'])
        self.assertIsInstance(fetched, dict)
        self.assertDictEqual(created, fetched)

    def test_execution_update(self):
        created = db_api.execution_create(EXECUTIONS[0]['workbook_name'],
                                          EXECUTIONS[0])
        self.assertIsInstance(created, dict)

        updated = db_api.execution_update(created['id'],
                                          {'task': 'task10'})
        self.assertIsInstance(updated, dict)
        self.assertEqual('task10', updated['task'])

        fetched = db_api.execution_get(created['id'])
        self.assertDictEqual(updated, fetched)

    def test_execution_list(self):
        created0 = db_api.execution_create(EXECUTIONS[0]['workbook_name'],
                                           EXECUTIONS[0])
        created1 = db_api.execution_create(EXECUTIONS[1]['workbook_name'],
                                           EXECUTIONS[1])

        fetched = db_api.executions_get(
            workbook_name=EXECUTIONS[0]['workbook_name'])

        self.assertEqual(2, len(fetched))
        self.assertDictEqual(created0, fetched[0])
        self.assertDictEqual(created1, fetched[1])

    def test_execution_delete(self):
        created = db_api.execution_create(EXECUTIONS[0]['workbook_name'],
                                          EXECUTIONS[0])
        self.assertIsInstance(created, dict)

        fetched = db_api.execution_get(created['id'])
        self.assertIsInstance(fetched, dict)
        self.assertDictEqual(created, fetched)

        db_api.execution_delete(created['id'])
        self.assertRaises(exc.NotFoundException,
                          db_api.execution_get,
                          created['id'])


TASKS = [
    {
        'id': '1',
        'workbook_name': 'my_workbook',
        'execution_id': '1',
        'name': 'my_task1',
        'description': 'my description',
        'requires': ['my_task2', 'my_task3'],
        'task_spec': None,
        'action_spec': None,
        'action': {'name': 'Nova:create-vm'},
        'state': 'IDLE',
        'tags': ['deployment'],
        'updated_at': None,
        'in_context': None,
        'parameters': None,
        'output': None,
        'task_runtime_context': None
    },
    {
        'id': '2',
        'workbook_name': 'my_workbook',
        'execution_id': '1',
        'name': 'my_task2',
        'description': 'my description',
        'requires': ['my_task4', 'my_task5'],
        'task_spec': None,
        'action_spec': None,
        'action': {'name': 'Cinder:create-volume'},
        'state': 'IDLE',
        'tags': ['deployment'],
        'updated_at': None,
        'in_context': {'image_id': '123123'},
        'parameters': {'image_id': '123123'},
        'output': {'vm_id': '343123'},
        'task_runtime_context': None
    },
]


class TaskTest(test_base.DbTestCase):
    def test_task_create_and_get(self):
        created = db_api.task_create(TASKS[0]['execution_id'],
                                     TASKS[0])
        self.assertIsInstance(created, dict)

        fetched = db_api.task_get(created['id'])
        self.assertIsInstance(fetched, dict)
        self.assertDictEqual(created, fetched)

    def test_task_update(self):
        created = db_api.task_create(TASKS[0]['execution_id'],
                                     TASKS[0])
        self.assertIsInstance(created, dict)

        updated = db_api.task_update(created['id'],
                                     {'description': 'my new desc'})
        self.assertIsInstance(updated, dict)
        self.assertEqual('my new desc', updated['description'])

        fetched = db_api.task_get(created['id'])
        self.assertDictEqual(updated, fetched)

    def test_task_list(self):
        created0 = db_api.task_create(TASKS[0]['execution_id'],
                                      TASKS[0])
        created1 = db_api.task_create(TASKS[1]['execution_id'],
                                      TASKS[1])

        fetched = db_api.tasks_get(
            workbook_name=TASKS[0]['workbook_name'])

        self.assertEqual(2, len(fetched))
        self.assertDictEqual(created0, fetched[0])
        self.assertDictEqual(created1, fetched[1])

    def test_task_delete(self):
        created = db_api.task_create(TASKS[0]['execution_id'],
                                     TASKS[0])
        self.assertIsInstance(created, dict)

        fetched = db_api.task_get(created['id'])
        self.assertIsInstance(fetched, dict)
        self.assertDictEqual(created, fetched)

        db_api.task_delete(created['id'])
        self.assertRaises(exc.NotFoundException, db_api.task_get,
                          created['id'])


class TXTest(test_base.DbTestCase):
    def test_rollback(self):
        db_api.start_tx()

        try:
            created = db_api.trigger_create(TRIGGERS[0])
            self.assertIsInstance(created, dict)

            fetched = db_api.trigger_get(created['id'])
            self.assertIsInstance(fetched, dict)
            self.assertDictEqual(created, fetched)

            self.assertTrue(self.is_db_session_open())

            db_api.rollback_tx()
        finally:
            db_api.end_tx()

        self.assertFalse(self.is_db_session_open())

        self.assertRaises(exc.NotFoundException,
                          db_api.trigger_get, created['id'])

        self.assertFalse(self.is_db_session_open())

    def test_commit(self):
        db_api.start_tx()

        try:
            created = db_api.trigger_create(TRIGGERS[0])
            self.assertIsInstance(created, dict)

            fetched = db_api.trigger_get(created['id'])
            self.assertIsInstance(fetched, dict)
            self.assertDictEqual(created, fetched)

            self.assertTrue(self.is_db_session_open())

            db_api.commit_tx()
        finally:
            db_api.end_tx()

        self.assertFalse(self.is_db_session_open())

        fetched = db_api.trigger_get(created['id'])
        self.assertIsInstance(fetched, dict)
        self.assertDictEqual(created, fetched)

        self.assertFalse(self.is_db_session_open())

    def test_rollback_multiple_objects(self):
        db_api.start_tx()

        try:
            created_trigger = db_api.trigger_create(TRIGGERS[0])
            self.assertIsInstance(created_trigger, dict)

            fetched_trigger = db_api.trigger_get(created_trigger['id'])
            self.assertIsInstance(fetched_trigger, dict)
            self.assertDictEqual(created_trigger, fetched_trigger)

            created_workbook = db_api.workbook_create(WORKBOOKS[0])
            self.assertIsInstance(created_workbook, dict)

            fetched_workbook = db_api.workbook_get(created_workbook['name'])
            self.assertIsInstance(fetched_workbook, dict)
            self.assertDictEqual(created_workbook, fetched_workbook)

            self.assertTrue(self.is_db_session_open())

            db_api.rollback_tx()
        finally:
            db_api.end_tx()

        self.assertFalse(self.is_db_session_open())

        self.assertRaises(exc.NotFoundException,
                          db_api.trigger_get, created_trigger['id'])

        self.assertRaises(exc.NotFoundException, db_api.workbook_get,
                          created_workbook['name'])

        self.assertFalse(self.is_db_session_open())

    def test_commit_multiple_objects(self):
        db_api.start_tx()

        try:
            created_trigger = db_api.trigger_create(TRIGGERS[0])
            self.assertIsInstance(created_trigger, dict)

            fetched_trigger = db_api.trigger_get(created_trigger['id'])
            self.assertIsInstance(fetched_trigger, dict)
            self.assertDictEqual(created_trigger, fetched_trigger)

            created_workbook = db_api.workbook_create(WORKBOOKS[0])
            self.assertIsInstance(created_workbook, dict)

            fetched_workbook = db_api.workbook_get(created_workbook['name'])
            self.assertIsInstance(fetched_workbook, dict)
            self.assertDictEqual(created_workbook, fetched_workbook)

            self.assertTrue(self.is_db_session_open())

            db_api.commit_tx()
        finally:
            db_api.end_tx()

        self.assertFalse(self.is_db_session_open())

        fetched_trigger = db_api.trigger_get(created_trigger['id'])
        self.assertIsInstance(fetched_trigger, dict)
        self.assertDictEqual(created_trigger, fetched_trigger)

        fetched_workbook = db_api.workbook_get(created_workbook['name'])
        self.assertIsInstance(fetched_workbook, dict)
        self.assertDictEqual(created_workbook, fetched_workbook)

        self.assertFalse(self.is_db_session_open())



# -*- coding: utf-8 -*-
#
# Copyright 2013 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

SAMPLE_TASK_SPEC = {
    'action': 'MyRest:create-vm',
    'name': 'create-vms',
    'on-success': ["format-volumes", {'task1': 'expression'}, {'task2': ''}],
    'on-finish': "attach-volumes",
    'on-error': ["task1", "task2"]
}

from mistral.tests import base
from mistral.workbook import tasks


class GetOnStateTest(base.BaseTest):
    def setUp(self):
        super(GetOnStateTest, self).setUp()
        self.task = tasks.TaskSpec(SAMPLE_TASK_SPEC)

    def test_state_finish(self):
        on_finish = self.task.get_on_finish()
        self.assertIsInstance(on_finish, dict)
        self.assertIn("attach-volumes", on_finish)

    def test_state_error(self):
        on_error = self.task.get_on_error()
        self.assertIsInstance(on_error, dict)
        self.assertEqual(len(on_error), 2)
        self.assertIn("task1", on_error)

    def test_state_success(self):
        on_success = self.task.get_on_success()
        self.assertIsInstance(on_success, dict)
        self.assertEqual(len(on_success), 3)
        self.assertIn("task1", on_success)
        self.assertIsNotNone(on_success["task1"])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import sys

import eventlet

eventlet.monkey_patch(
    os=True,
    select=True,
    socket=True,
    thread=False if '--use-debugger' in sys.argv else True,
    time=True)

import os

# If ../mistral/__init__.py exists, add ../ to Python search path, so that
# it will override what happens to be installed in /usr/(local/)lib/python...
POSSIBLE_TOPDIR = os.path.normpath(os.path.join(os.path.abspath(sys.argv[0]),
                                   os.pardir,
                                   os.pardir))
if os.path.exists(os.path.join(POSSIBLE_TOPDIR, 'mistral', '__init__.py')):
    sys.path.insert(0, POSSIBLE_TOPDIR)

from oslo.config import cfg
from oslo import messaging
from wsgiref import simple_server

from mistral.api import app
from mistral import config
from mistral import context
from mistral import engine
from mistral.engine import executor
from mistral.openstack.common import log as logging


LOG = logging.getLogger(__name__)


def launch_executor(transport):
    serializer = context.RpcContextSerializer(context.JsonPayloadSerializer())
    target = messaging.Target(topic=cfg.CONF.executor.topic,
                              server=cfg.CONF.executor.host)
    # Since engine and executor are tightly coupled, use the engine
    # configuration to decide which executor to get.
    endpoints = [executor.get_executor(cfg.CONF.engine.engine, transport)]
    server = messaging.get_rpc_server(
        transport, target, endpoints, executor='eventlet',
        serializer=serializer)
    server.start()
    server.wait()


def launch_engine(transport):
    serializer = context.RpcContextSerializer(context.JsonPayloadSerializer())
    target = messaging.Target(topic=cfg.CONF.engine.topic,
                              server=cfg.CONF.engine.host)
    endpoints = [engine.get_engine(cfg.CONF.engine.engine, transport)]
    server = messaging.get_rpc_server(
        transport, target, endpoints, executor='eventlet',
        serializer=serializer)
    server.start()
    server.wait()


def launch_api(transport):
    host = cfg.CONF.api.host
    port = cfg.CONF.api.port
    server = simple_server.make_server(host, port,
                                       app.setup_app(transport=transport))
    LOG.info("Mistral API is serving on http://%s:%s (PID=%s)" %
             (host, port, os.getpid()))
    server.serve_forever()


def launch_any(transport, options):
    # Launch the servers on different threads.
    threads = [eventlet.spawn(LAUNCH_OPTIONS[option], transport)
               for option in options]
    [thread.wait() for thread in threads]


# Map cli options to appropriate functions. The cli options are
# registered in mistral's config.py.
LAUNCH_OPTIONS = {
    'api': launch_api,
    'engine': launch_engine,
    'executor': launch_executor
}


def main():
    try:
        config.parse_args()
        logging.setup('Mistral')

        # Please refer to the oslo.messaging documentation for transport
        # configuration. The default transport for oslo.messaging is
        # rabbitMQ. The available transport drivers are listed in the
        # setup.cfg file in oslo.messaging under the entry_points section for
        # oslo.messaging.drivers. The transport driver is specified using the
        # rpc_backend option in the default section of the oslo configuration
        # file. The expected value for the rpc_backend is one of the key
        # values available for the oslo.messaging.drivers (i.e. rabbit, fake).
        # There are additional options such as ssl and credential that can be
        # specified depending on the driver.  Please refer to the driver
        # implementation for those additional options. It's important to note
        # that the "fake" transport should only be used if "all" the Mistral
        # servers are launched on the same process. Otherwise, messages do not
        # get delivered if the Mistral servers are launched on different
        # processes because the "fake" transport is using an in process queue.
        transport = messaging.get_transport(cfg.CONF)

        if (cfg.CONF.server == ['all']):
            # Launch all servers.
            launch_any(transport, LAUNCH_OPTIONS.keys())
        else:
            # Validate launch option.
            if set(cfg.CONF.server) - set(LAUNCH_OPTIONS.keys()):
                raise Exception('Valid options are all or any combination of '
                                'api, engine, and executor.')

            # Launch distinct set of server(s).
            launch_any(transport, set(cfg.CONF.server))

    except RuntimeError as excp:
        sys.stderr.write("ERROR: %s\n" % excp)
        sys.exit(1)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
#
# Copyright 2013 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from pecan import rest
from wsme import types as wtypes
import wsmeext.pecan as wsme_pecan

from mistral.api.controllers import resource
from mistral.api.controllers.v1 import execution
from mistral.api.controllers.v1 import listener
from mistral.api.controllers.v1 import workbook_definition
from mistral.db import api as db_api
from mistral.openstack.common import log as logging
from mistral.services import workbooks
from mistral.utils import rest_utils

LOG = logging.getLogger(__name__)
SCOPE_TYPES = wtypes.Enum(str, 'private', 'public')


class Workbook(resource.Resource):
    """Workbook resource."""

    name = wtypes.text
    description = wtypes.text
    tags = [wtypes.text]
    scope = SCOPE_TYPES

    @classmethod
    def sample(cls):
        return cls(name='flow',
                   description='my workflow',
                   tags=['large', 'expensive'])


class Workbooks(resource.Resource):
    """A collection of Workbooks."""

    workbooks = [Workbook]

    @classmethod
    def sample(cls):
        return cls(workbooks=[Workbook.sample()])


class WorkbooksController(rest.RestController):
    definition = workbook_definition.WorkbookDefinitionController()
    listeners = listener.ListenersController()
    executions = execution.WorkbookExecutionsController()

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(Workbook, wtypes.text)
    def get(self, name):
        """Return the named workbook."""
        LOG.debug("Fetch workbook [name=%s]" % name)

        values = db_api.workbook_get(name)
        return Workbook.from_dict(values)

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(Workbook, wtypes.text, body=Workbook)
    def put(self, name, workbook):
        """Update the named workbook."""
        LOG.debug("Update workbook [name=%s, workbook=%s]" % (name, workbook))

        return Workbook.from_dict(db_api.workbook_update(name,
                                                         workbook.to_dict()))

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(Workbook, body=Workbook, status_code=201)
    def post(self, workbook):
        """Create a new workbook."""
        LOG.debug("Create workbook [workbook=%s]" % workbook)
        wb = workbooks.create_workbook(workbook.to_dict())
        return Workbook.from_dict(wb)

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(None, wtypes.text, status_code=204)
    def delete(self, name):
        """Delete the named workbook."""
        LOG.debug("Delete workbook [name=%s]" % name)
        db_api.workbook_delete(name)

    @wsme_pecan.wsexpose(Workbooks)
    def get_all(self):
        """return all workbooks.

        Where project_id is the same as the requestor or
        project_id is different but the scope is public.
        """
        LOG.debug("Fetch workbooks.")

        workbooks_list = [Workbook.from_dict(values)
                          for values in db_api.workbooks_get()]

        return Workbooks(workbooks=workbooks_list)



# -*- coding: utf-8 -*-
#
# Copyright 2013 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Valid task and workflow states."""

IDLE = 'IDLE'
RUNNING = 'RUNNING'
SUCCESS = 'SUCCESS'
ERROR = 'ERROR'
STOPPED = 'STOPPED'
DELAYED = 'DELAYED'

_ALL = [IDLE, RUNNING, SUCCESS, ERROR, STOPPED, DELAYED]


def is_valid(state):
    return state in _ALL


def is_finished(state):
    return state in [SUCCESS, ERROR]


def is_stopped_or_finished(state):
    return state == STOPPED or is_finished(state)
